--------------------------------------------------------------------------------
-- 08-Apr-1997 : Chas. : Creation.
-- 22-Apr-1997 : Chas. : Added role access_admin
-- 07-May-1997 : Chas. : Added ChangePass procedure.
-- 07-May-1997 : Chas. : Added RAISE_APPLICATION_ERROR logic
--------------------------------------------------------------------------------
set echo on
set termout on

@@cr_seq_audit_admin
@@cr_audit_admin

spool useradmin.log

DROP PUBLIC SYNONYM UserAdmin
/

CREATE OR REPLACE PACKAGE UserAdmin IS
	PROCEDURE CreateUser(	IN_Username VARCHAR2,
				IN_Password VARCHAR2 );
	PROCEDURE DropUser(	IN_Username VARCHAR2 );
	PROCEDURE ChangePass(	IN_Username VARCHAR2,
				IN_Password VARCHAR2 );
	PROCEDURE GrantRole(	IN_Username VARCHAR2,
				IN_Role VARCHAR2,
				IN_DefaultYN VARCHAR2 DEFAULT 'N' );
	PROCEDURE RevokeRole(	IN_Username VARCHAR2,
				IN_Role VARCHAR2 );
END UserAdmin;
/

CREATE OR REPLACE PACKAGE BODY  UserAdmin IS

--------------------------------------------------------------------------------
-- Note:	This package should be owned by SYSTEM because it must be owned
--		by an id that has CREATE SESSION privs (schema ID's should NOT)
--		Also, SYS must make the following explicit grants to SYSTEM:
--			GRANT ALTER USER TO system
--			GRANT CREATE USER TO system
--			GRANT CREATE SESSION TO system WITH ADMIN OPTION
--			GRANT DROP USER TO system
--			GRANT GRANT ANY ROLE TO system
--			GRANT SELECT ON DBA_ROLE_PRIVS TO system
--			GRANT SELECT ON DBA_SEGMENTS TO system
--		Optional Grants:
--			GRANT SELECT ON dba_role_privs TO access_admin
--			GRANT SELECT ON dba_roles      TO access_admin
--------------------------------------------------------------------------------

PROCEDURE AuditUserAdmin(	IN_Proc		VARCHAR2,
				IN_Info		VARCHAR2 DEFAULT NULL,
				IN_Error	NUMBER	DEFAULT NULL ) IS
BEGIN
	INSERT INTO audit_admin
		(	audit_id,
			procname,
			info,
			errornum,
			audit_user,
			audit_date
		)
	VALUES (	seq_audit_admin.nextval,
			IN_Proc,
			IN_info,
			IN_Error,
			user,
			sysdate
		);

	COMMIT;
END AuditUserAdmin;

--------------------------------------------------------------------------------
-- PROCEDURE CreateUser creates a user and grants CREATE SESSION to him/her
--------------------------------------------------------------------------------
PROCEDURE CreateUser( IN_Username VARCHAR2, IN_Password VARCHAR2 ) IS
hCursor		NUMBER;
vProcName	audit_admin.procname%TYPE	:= 'CreateUser';
vInfo		audit_admin.info%TYPE		:= NULL;
BadPassword	EXCEPTION;
UserExists	EXCEPTION;
NameTooLong	EXCEPTION;

BEGIN
	IF UPPER(IN_Username) = UPPER(IN_Password)
	THEN
		RAISE BadPassword;
	END IF;

	IF length(IN_Username) > 8
	THEN
		RAISE NameTooLong;
	END IF;

	hCursor := dbms_sql.open_cursor;
	dbms_sql.parse(	hCursor,
			'CREATE USER '		|| IN_Username	||
			' IDENTIFIED BY '	|| IN_Password	||
			' DEFAULT TABLESPACE users TEMPORARY TABLESPACE temp',
			dbms_sql.v7);
	dbms_sql.close_cursor( hCursor);

	hCursor := dbms_sql.open_cursor;
	dbms_sql.parse(	hCursor,
			'GRANT CREATE SESSION TO ' || IN_Username,
			dbms_sql.v7);
	dbms_sql.close_cursor( hCursor);

	vInfo := 'Created ' || IN_Username || ' Password ' || IN_Password;
	AuditUserAdmin( vProcName, vInfo );

EXCEPTION
	WHEN NameTooLong THEN
		vInfo := '!Bad Username: ' || IN_Username;
		AuditUserAdmin( vProcName, vInfo );
		dbms_output.put_line( 'Username must be 8 chars or less.');
		RAISE_APPLICATION_ERROR(-20010, 'CreateUser:Username too long');
	WHEN BadPassword THEN
		vInfo := '!Bad Password for ' || IN_Username;
		AuditUserAdmin( vProcName, vInfo );
		dbms_output.put_line( 'Username and Password must differ.');
		RAISE_APPLICATION_ERROR(-20020, 'CreateUser:Username=Password');
	WHEN OTHERS THEN
		vInfo := '!Create User ' || IN_Username;
		AuditUserAdmin( vProcName, vInfo , sqlcode);
		dbms_output.put_line('error ' || sqlerrm);
		RAISE_APPLICATION_ERROR(-20030, 'CreateUser:Error ' || sqlerrm);
END CreateUser;

--------------------------------------------------------------------------------
-- PROCEDURE DropUser Drops a user.
--------------------------------------------------------------------------------
PROCEDURE DropUser( IN_Username VARCHAR2 ) IS
hCursor		NUMBER;
vProcName	audit_admin.procname%TYPE	:= 'DropUser';
vInfo		audit_admin.info%TYPE		:= NULL;
vSegCount	NUMBER;
UserHasObjects	EXCEPTION;
UserIsDBA	EXCEPTION;
BEGIN
	IF upper( IN_Username ) IN ('SYS', 'SYSTEM')
	THEN
		RAISE UserIsDBA;
	END IF;

	SELECT	count(*)
	INTO	vSegCount
	FROM	dba_segments
	WHERE	owner = UPPER( IN_Username );

	IF vSegCount > 0
	THEN
		RAISE UserHasObjects;
	END IF;

	hCursor := dbms_sql.open_cursor;
	dbms_sql.parse(	hCursor,
			'DROP USER ' || IN_Username || ' CASCADE',
			dbms_sql.v7);
	dbms_sql.close_cursor( hCursor);

	vInfo := 'Dropped ' || IN_Username;
	AuditUserAdmin( vProcName, vInfo );

EXCEPTION
	WHEN UserIsDBA THEN
		vInfo := '!DROP ' || IN_Username;
		AuditUserAdmin( vProcName, vInfo );
		dbms_output.put_line('Cannot drop SYS or SYSTEM accounts.');
		RAISE_APPLICATION_ERROR(-2110, 'DropUser: User is DBA');
	WHEN UserHasObjects THEN
		vInfo := '!DROP ' || IN_Username;
		AuditUserAdmin( vProcName, vInfo );
		RAISE_APPLICATION_ERROR(-20120, 'DropUser: User owns objects');
	WHEN OTHERS THEN
		vInfo := '!DROP ' || IN_Username;
		AuditUserAdmin( vProcName, vInfo, sqlcode );
		dbms_output.put_line('error ' || sqlerrm);
		RAISE_APPLICATION_ERROR(-20130, 'DropUser: Error ' || sqlerrm);
END DropUser;

--------------------------------------------------------------------------------
-- PROCEDURE ChangePass changes a user's password
--------------------------------------------------------------------------------
PROCEDURE ChangePass(	IN_Username VARCHAR2, IN_Password VARCHAR2 ) IS
hCursor			NUMBER;
vProcName		audit_admin.procname%TYPE	:= 'ChangePass';
vInfo			audit_admin.info%TYPE		:= NULL;
BadPassword		EXCEPTION;
PassTooLong		EXCEPTION;
UserIsDBA		EXCEPTION;
BEGIN
	IF upper( IN_Username ) IN ('SYS', 'SYSTEM')
	THEN
		RAISE UserIsDBA;
	END IF;

	IF UPPER(IN_Username) = UPPER(IN_Password)
	THEN
		RAISE BadPassword;
	END IF;

	IF length(IN_Password) > 30
	THEN
		RAISE PassTooLong;
	END IF;

	hCursor := dbms_sql.open_cursor;
	dbms_sql.parse(	hCursor,
			'ALTER USER '		|| IN_Username	||
			' IDENTIFIED BY '	|| IN_Password,
			dbms_sql.v7);
	dbms_sql.close_cursor( hCursor);

	vInfo := 'Changed ' || IN_Username || ' Password ' || IN_Password;
	AuditUserAdmin( vProcName, vInfo );

EXCEPTION
	WHEN UserIsDBA THEN
		vInfo := '!DROP ' || IN_Username;
		AuditUserAdmin( vProcName, vInfo );
		dbms_output.put_line('Cannot change SYS or SYSTEM passwords.');
		RAISE_APPLICATION_ERROR(-20210, 'ChangePass:Cannot change DBA');
	WHEN PassTooLong THEN
		vInfo := '!Bad Username: ' || IN_Username;
		AuditUserAdmin( vProcName, vInfo );
		dbms_output.put_line( 'Password must be 30 chars or less.');
		RAISE_APPLICATION_ERROR(-20220, 'ChangePass:password>30 chars');
	WHEN BadPassword THEN
		vInfo := '!Bad Password for ' || IN_Username;
		AuditUserAdmin( vProcName, vInfo );
		dbms_output.put_line( 'Username and Password must differ.');
		RAISE_APPLICATION_ERROR(-20230, 'ChangePass:username=password');
	WHEN OTHERS THEN
		vInfo := '!Create User ' || IN_Username;
		AuditUserAdmin( vProcName, vInfo , sqlcode);
		RAISE_APPLICATION_ERROR(-20240, 'ChangePass:Error ' || sqlerrm);
END ChangePass;
--------------------------------------------------------------------------------
-- PROCEDURE GrantRole grants roles
--------------------------------------------------------------------------------
PROCEDURE GrantRole(	IN_Username VARCHAR2,
			IN_Role VARCHAR2,
			IN_DefaultYN VARCHAR2 DEFAULT 'N') IS
CURSOR cDefRole IS
	SELECT	granted_role
	FROM	dba_role_privs
	WHERE	grantee = UPPER(IN_Username)
	AND	default_role = 'YES';

hCursor		NUMBER;
vProcName	audit_admin.procname%TYPE	:= 'GrantRole';
vInfo		audit_admin.info%TYPE		:= NULL;
vRoleStr	VARCHAR2(2000)	:= NULL;
vDefRoleCount	NUMBER		:= 0;
vHasDefRoles	BOOLEAN		:= FALSE;
vRoleCheck	NUMBER		:= 0;
UserIsDBA	EXCEPTION;
NoPrivRoles	EXCEPTION;
RoleAlready	EXCEPTION;
BEGIN
	vInfo := 'Grant ' || IN_Role || ' to ' || IN_Username;
	
	IF UPPER(IN_Username) IN ('SYS', 'SYSTEM')
	THEN
		RAISE UserIsDBA;
	END IF;

	IF UPPER(IN_Role) IN (	'DBA', 'CONNECT', 'RESOURCE')
	THEN
		RAISE NoPrivRoles;
	END IF;

	SELECT	count(*)
	INTO	vRoleCheck
	FROM	dba_role_privs
	WHERE	grantee		= UPPER(IN_Username)
	AND	granted_role	= UPPER(IN_Role);

	IF vRoleCheck > 0
	THEN
		RAISE RoleAlready;
	END IF;

	FOR rDefRole IN cDefRole LOOP
		vRoleStr	:= vRoleStr || rDefRole.granted_role || ',';
		vDefRoleCount	:= vDefRoleCount + 1;
	END LOOP;

	IF ( vDefRoleCount > 0 ) OR ( IN_DefaultYN = 'Y' )
	THEN
		vHasDefRoles := TRUE;
		IF ( vDefRoleCount > 0 ) AND ( IN_DefaultYN = 'N' )
		THEN
			vRoleStr := substr(vRoleStr, 1, length(vRoleStr) - 1 );
		ELSIF ( vDefRoleCount > 0 ) AND ( IN_DefaultYN = 'Y' )
			THEN
				vRoleStr := vRoleStr || ' ' || IN_Role;
		ELSIF ( vDefRoleCount = 0 ) AND ( IN_DefaultYN = 'Y' )
			THEN
				vRoleStr := IN_Role;
		END IF;
	END IF;
------------------------------
-- First grant the new role --
------------------------------
	hCursor := dbms_sql.open_cursor;
	dbms_sql.parse(	hCursor,
			'GRANT ' || IN_Role || ' TO '|| IN_Username,
			dbms_sql.v7);
	dbms_sql.close_cursor( hCursor);

--------------------------------------
-- Now set user to no default roles --
--------------------------------------
	hCursor := dbms_sql.open_cursor;
	dbms_sql.parse(	hCursor,
			'ALTER USER ' || IN_Username || ' DEFAULT ROLE NONE',
			dbms_sql.v7);
	dbms_sql.close_cursor( hCursor);

---------------------------------
-- Now grant any default roles --
---------------------------------
	IF vHasDefRoles = TRUE
	THEN
		hCursor := dbms_sql.open_cursor;
		dbms_sql.parse(	hCursor,
			'ALTER USER ' ||IN_Username||' DEFAULT ROLE '||vRoleStr,
			dbms_sql.v7);
		dbms_sql.close_cursor( hCursor);
	
	END IF;

	AuditUserAdmin( vProcName, vInfo );

EXCEPTION
	WHEN UserIsDBA THEN
		vInfo := '!' || vInfo;
		AuditUserAdmin( vProcName, vInfo );
		dbms_output.put_line('Cannot grant to SYS or SYSTEM accounts.');
		RAISE_APPLICATION_ERROR(-2410, 'GrantRole: Cannot grant SYS');
	WHEN RoleAlready THEN
		vInfo := '!' || vInfo;
		AuditUserAdmin( vProcName, vInfo );
		dbms_output.put_line(IN_Username ||' already has ' || IN_Role);
		RAISE_APPLICATION_ERROR(-20420, 'GrantRole:Role already grntd');
	WHEN NoPrivRoles THEN
		vInfo := '!' || vInfo;
		AuditUserAdmin( vProcName, vInfo );
		dbms_output.put_line('Not authorized to grant ' || IN_Role);
		RAISE_APPLICATION_ERROR(-20430, 'GrantRole: Cannot grant DBA');
	WHEN OTHERS THEN
		vInfo := '!' || vInfo;
		AuditUserAdmin( vProcName, vInfo, sqlcode );
		dbms_output.put_line('error ' || sqlerrm);
		RAISE_APPLICATION_ERROR(-20440, 'GrantRole: Error ' || sqlerrm);

END GrantRole;

--------------------------------------------------------------------------------
-- PROCEDURE RevokeRole  revokes roles
--------------------------------------------------------------------------------
PROCEDURE RevokeRole( IN_Username VARCHAR2, IN_Role VARCHAR2 ) IS
hCursor		NUMBER;
vProcName	audit_admin.procname%TYPE	:= 'RevokeRole';
vInfo		audit_admin.info%TYPE		:= NULL;
UserIsDBA	EXCEPTION;

BEGIN
	IF UPPER(IN_Username) IN ('SYS', 'SYSTEM')
	THEN
		RAISE UserIsDBA;
	END IF;

	hCursor := dbms_sql.open_cursor;
	dbms_sql.parse(	hCursor,
			'REVOKE ' || IN_Role || ' FROM ' || IN_Username,
			dbms_sql.v7);
	dbms_sql.close_cursor( hCursor);

	vInfo := 'Revoke ' || IN_Role || ' from ' || IN_Username;
	AuditUserAdmin( vProcName, vInfo );

EXCEPTION
	WHEN UserIsDBA THEN
		vInfo := '!Revoke ' || IN_Role || ' FROM ' || IN_Username;
		AuditUserAdmin( vProcName, vInfo );
		dbms_output.put_line('Cannot revoke from SYS or SYSTEM.');
		RAISE_APPLICATION_ERROR(-20510, 'RevokeRole:Cannot revoke SYS');
	WHEN OTHERS THEN
		vInfo := '!Revoke ' || IN_Role || ' FROM ' || IN_Username;
		AuditUserAdmin( vProcName, vInfo, sqlcode );
		dbms_output.put_line('error ' || sqlerrm);
		RAISE_APPLICATION_ERROR(-20520, 'RevokeRole:Error ' || sqlerrm);

END RevokeRole;

END UserAdmin;
/

CREATE PUBLIC SYNONYM UserAdmin FOR UserAdmin
/
CREATE ROLE access_admin
/
GRANT EXECUTE ON useradmin TO access_admin
/

spool off
